import pysam
import sys
from searcHPV.general import *
import re

##################
#find the end of each cigar string slot
#pos: start position of each cigar string slot
#cigar: cigar string returned by pysam

def find_end(pos,cigar):
    letters = re.split('\d',cigar)
    letters = list(filter(None, letters))
    numbers = re.split('\D',cigar)
    numbers = list(filter(None, numbers))
    numbers = list(map(int,numbers))
    length = 0
    if letters[0] == 'S' or letters[0] == 'H':
        return int(pos)
    elif letters[0] == 'M':
        for i,letter in enumerate(letters):
            if letter == 'M':
                length+=numbers[i]
            elif letter == 'D':
                length+=numbers[i]
            elif letter == 'I':
                length-=numbers[i]
            else:
                continue
        return int(pos)+length

#####################
#identify fusion points on human genome
#Function: identify fusion points on human genome by PE and SP reads
#bam: alignment.RG.indelre.mkdup.sort.bam
#virus_chrm: name of virus chromosome in virus reference genome
##out_dir:output path
#return: output file
def define_fusion(bam,virus_chrm,out_dir):
    out_dir = os.path.abspath(out_dir)
    paired_read_li = []
    clipped_read_li = []

    samfile = pysam.AlignmentFile(bam, 'rb')
    #print(samfile.references)
    if samfile.count(virus_chrm) != 0:
        for read in samfile.fetch(virus_chrm):
            # some filter here
            if read.is_duplicate is True:
                continue
            if read.is_qcfail is True:
                continue
            if read.is_unmapped is True:
                continue
            if read.is_secondary is True:
                continue
            mapq = read.mapping_quality
            if mapq < 50:
                continue
            if read.next_reference_name != virus_chrm:
                paired_read_li.append(read)
            else:
                cigar = read.cigarstring
                if "H" in cigar:
                    clipped_read_li.append(read)
                    
                elif "S" in cigar:
                    clipped_read_li.append(read)
                    
                else:
                    continue

    # for each in paired_read_li:
    #     if each.query_name == 'HWI-ST1129:195:C0YAUACXX:5:1205:4558:86030':
    #         print(each.get_tag('SA'),each.reference_name,each.next_reference_name)

    next_pos_li = []
    for read in paired_read_li:
        next_pos_li.append(f"{read.next_reference_name}:{read.next_reference_start}")
    next_pos_li_sort = sorted(next_pos_li)
    
    from operator import itemgetter
    clipped_infor_li = []
    for read in clipped_read_li:
        if read.is_duplicate is True:
            continue
        if read.is_qcfail is True:
            continue
        if read.is_unmapped is True:
            continue
        if read.is_secondary is True:
            continue
        if read.has_tag('SA'):
            name = read.query_name
            SATag = read.get_tag('SA')
            nchrm = SATag.split(",")[0]
            npos = SATag.split(",")[1]
            ndir = SATag.split(",")[2]
            ncigar = SATag.split(",")[3]
            mapq = int(SATag.split(",")[4])
            if mapq >= 50:
                clipped_infor_li.append([nchrm,npos,ndir,ncigar,name])
                # if nchrm == '10':
                #     print(read,samfile.mate(read))

    #10:7041919:11:1;17:41153083:47:99;17:41153186:12:99;2:102871348:6:1;2:17687921:8:1;3:130887923:2:8;5:97050065:7:1;6:151124993:11:1;6:2255928:8:1;6:71772584:8:1;X:113760995:6:1
        else:
            continue
    sorted_clipped_infor_li = sorted(clipped_infor_li,key=itemgetter(0,1))
    #print(sorted_clipped_infor_li)
    pos_li = []
    for read in sorted_clipped_infor_li:
        chrm = read[0]
        if chrm == virus_chrm:
            continue
        else:
            end_pos = find_end(read[1],read[3])
            pos_li.append((chrm,end_pos))
            

    a = set(pos_li)
    #print(pos_li)
    b = []
    for i in a:
        count = pos_li.count(i)
        i += (str(count),)
        b.append(i)
    b = sorted(b)

    for i,pos_infor in enumerate(b):
        chrm = pos_infor[0]
        pos = int(pos_infor[1])
        samfile = pysam.AlignmentFile(bam, 'rb')
        paired_evidence_count = 0
        if pos < 150:
            start_pos = 1
        else:
            start_pos = pos -150
        if samfile.count(chrm,start_pos,pos+150) != 0:
            for read in samfile.fetch(chrm,start_pos,pos+150):
                # some filter here
                if read.is_duplicate is True:
                    continue
                if read.is_qcfail is True:
                    continue
                if read.is_unmapped is True:
                    continue
                if read.is_secondary is True:
                    continue
                mapq = read.mapping_quality
                if mapq < 50:
                    continue
                if read.next_reference_name == virus_chrm:
                    paired_evidence_count += 1
        b[i] += (str(paired_evidence_count),)
    
    with open(f'{out_dir}/{virus_chrm}.genome_fusion.txt','w') as output:
        output.write('chrm\tpos\tsingle_evidence\tpaired_evidence\n')
        for i in b:
            output.write(f'{i[0]}\t{i[1]}\t{i[2]}\t{i[3]}\n')
    return f'{out_dir}/{virus_chrm}.genome_fusion.txt'

##################
#cluster fusion points within certain base pair
#resultin: genome_fusion.txt
#outf: output file
#window: length of window for cluster, default = 100
def cluster_result(resultin,outf,window=100):
    with open(resultin) as resultin:
        with open(outf,'w') as output:
            out_li = []
            for line in resultin.read().rstrip().split('\n'):
                elements = line.rstrip()
                if elements == "":
                    out_li.append(elements)
                else:
                    pos_li = elements.split(';')
                    pos_li = sorted(pos_li)
                    new_pos_li = [pos_li[0]]
                    for i,pos in enumerate(pos_li[1:]):
                        old_infor = new_pos_li[-1].split(':')
                        old_chrm = old_infor[0]
                        old_p = int(old_infor[1])
                        old_count = int(old_infor[2])
                        infor = pos.split(':')
                        chrm = infor[0]
                        p = int(infor[1])
                        single_count = int(infor[2])
                        if chrm == old_chrm:
                            if abs(p-old_p) > window:
                                new_pos_li.append(':'.join(infor))
                            else:
                                if single_count > old_count:
                                    new_pos_li[-1] = ':'.join(infor)
                        else:
                            new_pos_li.append(':'.join(list(map(str,infor))))
                    
                    pos_infor = ';'.join(new_pos_li)   
                    output.write(pos_infor)
    return None

    






import sys
import os
import re
import pysam
from searcHPV.general import *

########################
#get informative reads (SP + PE) from bam file
#chrm:chromosome on human genome
#pos: start position
#bam: original bam file
#windowSize: default = 300, length of window for extracting reads
#ref_genome:name of virus chromosome
def get_reads(chrm,pos,bam,ref_genome,windowSize=300):
    samfile = pysam.AlignmentFile(bam, "rb")
    read_name_list = []
    #print (f'{chrm}:{pos}')
    for read in samfile.fetch(chrm,pos-int(windowSize),pos+int(windowSize)):
        read_name = read.query_name    
        read_ref = read.reference_name
        pair_ref = read.next_reference_name

        if read.is_duplicate is True:
            continue
        if read.is_qcfail is True:
            continue
        if read.is_unmapped is True:
            continue
        if read.is_secondary is True:
            continue
#pair end reads
        if pair_ref == ref_genome:
            read_name_list.append(read_name)
            
            
#split reads
        else:
            if read.has_tag('SA'):
                name = read.query_name
                SATag = read.get_tag('SA')
                nchrm = SATag.split(",")[0]
                if nchrm == ref_genome:
                    npos = SATag.split(",")[1]
                    ndir = SATag.split(",")[2]
                    ncigar = SATag.split(",")[3]
                    read_name_list.append(read_name)
                    
    read_name_set = set(read_name_list)

    samfile.close()
    return read_name_set

########################
#Function:
#extract informative reads name from original bam file
#bam: original bam file, alignment.RG.indelre.mkdup.sort.bam
#fusionRes: genome fusion result file from genome fusion step, all.filtered.clustered.result
#out_dir:out put directory all
#window:the length of region searching for informative reads, default=300
def extract_read_name(bam,out_dir,virRef,window):
    #####for this step, need to modify from Yifan's script
    #candidate_in = os.popen(f'grep \'{sample}\' {out_dir}/call_fusion/all.filtered.clustered.result')
    #############
    virus_chrm_list = []
    with open(virRef) as virRefFile:
        for each in virRefFile.read().rstrip().split('\n'):
            if '>' in each:
                virus_chrm_list.append(each.split()[0].replace('>',''))
    for virus_chrm in virus_chrm_list:
        fusionRes = f'{out_dir}/call_fusion/{virus_chrm}.all.filtered.clustered.result'
        check_file(fusionRes)
        with open(fusionRes) as candidate_in:
            candidate_in = candidate_in.read().rstrip()

        #if there are fusion points
        if candidate_in != "":
            candidate_in = candidate_in.split(";")
#for each candidate site, collect reads with truncation and abberant pairs upstream and downstream windowbp and reads in HPV
    # sampleDir = f'/home/wenjingu/myScratch/ourResult/newVersion/targetedExomeResult/{sample}.{windowSize}'
    # try:
    #     subprocess.call(f'mkdir {sampleDir}')
    # except:
    #     print('dir exists')
    # try:
    #     subprocess.call(f'mkdir {sampleDir}/readName')
    # except:
    #     print('dir exists')
    #find the virus_chrm
        for site in candidate_in:
            chrm = site.split(':')[0]
            pos = int(site.split(':')[1])
            #print(chrm,pos,bam,virus_chrm)
            read_list = get_reads(chrm=chrm,pos=pos,bam=bam,ref_genome =virus_chrm,windowSize=window)
            outf_path = f'{out_dir}/assemble/{chrm}.{pos}/'
            mkdir(outf_path)
            outf = open(f'{outf_path}/readName.txt', 'a+')
        #my dir
            for each in read_list:
                print(f'>{each}',file=outf)
            outf.close()
    return None

########################
#Function:generate bash script for extracting read sequence
#out_dir:output directory for assemble
#fq1:raw sequenction data, fastq file
#fq2:raw sequencing data, fastq file
#return: path of bash script
#gz: if fastq file is in gz format
def extract_read_seq(out_dir,fq1,fq2,gz):
    with open(f'{out_dir}/extractReadSequence.sh','w') as bash:
        bash.write('#!/bin/bash\n')
        for site in os.listdir(out_dir):
            if ".sh" not in site:
                readNamePath = f'{out_dir}/{site}/readName.txt'
                outputPath = f'{out_dir}/{site}/supportiveReads/'
                if not os.path.exists(outputPath):
                    os.system(f'mkdir -p {outputPath}')
                tab = '\"\\t\"'
                newLine = '\"\\n\"'
                if gz:
                    bash.write(f'''zcat {fq1} | awk '{{if(NR%4!=0)ORS={tab};else ORS={newLine}}}1' | awk -F {tab} 'NR==FNR{{split($1,b," ");a[b[1]]=($1{newLine}$2{newLine}$3{newLine}$4)}}NR>FNR{{gsub(/>/,"@");if(a[$1]!={newLine})print a[$1]}}' - {readNamePath} | grep -v '^$' > {outputPath}/{site}.informativeReads.1.fq;
zcat {fq2} | awk '{{if(NR%4!=0)ORS={tab};else ORS={newLine}}}1' | awk -F {tab} 'NR==FNR{{split($1,b," ");a[b[1]]=($1{newLine}$2{newLine}$3{newLine}$4)}}NR>FNR{{gsub(/>/,"@");if(a[$1]!={newLine})print a[$1]}}' - {readNamePath} | grep -v '^$' > {outputPath}/{site}.informativeReads.2.fq;
''')
                else:
                    bash.write(f'''awk '{{if(NR%4!=0)ORS={tab};else ORS={newLine}}}1' {fq1} | awk -F {tab} 'NR==FNR{{split($1,b," ");a[b[1]]=($1{newLine}$2{newLine}$3{newLine}$4)}}NR>FNR{{gsub(/>/,"@");if(a[$1]!={newLine})print a[$1]}}' - {readNamePath} | grep -v '^$' > {outputPath}/{site}.informativeReads.1.fq;
awk '{{if(NR%4!=0)ORS={tab};else ORS={newLine}}}1' {fq2} | awk -F {tab} 'NR==FNR{{split($1,b," ");a[b[1]]=($1{newLine}$2{newLine}$3{newLine}$4)}}NR>FNR{{gsub(/>/,"@");if(a[$1]!={newLine})print a[$1]}}' - {readNamePath} | grep -v '^$' > {outputPath}/{site}.informativeReads.2.fq;
''')
        bash.write('echo \'extract informative sequences done\'')
    return f'{out_dir}/extractReadSequence.sh'

########################
#Function:formalize fasta file for PEAR
#out_dir:output directory for assemble
def preprocessForPear(out_dir):
    listSites = os.listdir(out_dir)
    for site in listSites:
        if '.sh' not in site:
            outputPath = f'{out_dir}/{site}/supportiveReads/'
            with open(f'{outputPath}/{site}.informativeReads.1.fq','r') as inputFile:
                with open(f'{outputPath}/{site}.formalizedReads.1.fq','w') as outputFile:
                    i = 0
                    readList = inputFile.read().split('\n')
                    for each in readList:
                        if '@' in each and '+' in readList[i+4]:
                            rowSeq = readList[i+1] + readList[i+2]+readList[i+3]
                            rowInfo = readList[i+5] + readList[i+6]+readList[i+7]
                            outputFile.write(each+'\n'+rowSeq+'\n+\n'+rowInfo+'\n')
                        elif '@' in each:
                            rowSeq = readList[i+1] + readList[i+2]+readList[i+3]
                            rowInfo = readList[i+1] + readList[i+2]+readList[i+3]
                            outputFile.write(each+'\n'+rowSeq+'\n+\n'+rowInfo+'\n')
                        i += 1

            with open(f'{outputPath}/{site}.informativeReads.2.fq','r') as inputFile:
                with open(f'{outputPath}/{site}.formalizedReads.2.fq','w') as outputFile:
                    i = 0
                    readList = inputFile.read().split('\n')
                    for each in readList:
                        if '@' in each and '+' in readList[i+4]:
                            rowSeq = readList[i+1] + readList[i+2]+readList[i+3]
                            rowInfo = readList[i+5] + readList[i+6]+readList[i+7]
                            outputFile.write(each+'\n'+rowSeq+'\n+\n'+rowInfo+'\n')
                        elif '@' in each:
                            rowSeq = readList[i+1] + readList[i+2]+readList[i+3]
                            rowInfo = readList[i+1] + readList[i+2]+readList[i+3]
                            outputFile.write(each+'\n'+rowSeq+'\n+\n'+rowInfo+'\n')
                        i += 1  
    return None

########################
#Function:generate bash script for PEAR
#out_dir:output directory for assemble
#return:path of bash script
def pear(out_dir):
    listSites = os.listdir(out_dir)
    with open(f'{out_dir}/pear.sh','w') as output:
        output.write('#!/bin/bash\n')
        for site in listSites:
            if ".sh" not in site:
                outputPath = f'{out_dir}/{site}/pearOutput/'
                fqPath = f'{out_dir}/{site}/supportiveReads/'
                if not os.path.exists(outputPath):
                    os.mkdir(outputPath)
                output.write(f'''
    pear \
    -f {fqPath}/{site}.informativeReads.1.fq \
    -r {fqPath}/{site}.informativeReads.2.fq \
    -o {outputPath}/{site}''')
    return f'{out_dir}/pear.sh'

#############
#Function: formalize results from PEAR for CAP3
#out_dir:output directory for assemble
def preprocessForCap3(out_dir):
    listSites = os.listdir(out_dir)
    for site in listSites:
        if ".sh" not in site:
            pearOutputPath = f'{out_dir}/{site}/pearOutput/'

            #change unassemble reads to be fasta
            with open(f'{pearOutputPath}/{site}.unassembled.forward.fastq',encoding='utf8',errors='ignore') as fq:
                with open(f'{pearOutputPath}/{site}.1.fa','w') as fa:
                    #fq = fq.read().rstrip().decode('uft-16').split('\n')
                    fq = fq.read().rstrip().split('\n')
                    if fq != ['']:
                        for i in range(len(fq)):
                            if i%4 == 0:
                                newName = '>' + fq[i][1:]
                                rowSeq = fq[i+1]
                                fa.write( newName + '.1' + '\n' + rowSeq + '\n')
                        
            with open(f'{pearOutputPath}/{site}.unassembled.reverse.fastq',encoding='utf8',errors='ignore') as fq:
                with open(f'{pearOutputPath}/{site}.2.fa','w') as fa:
                    #fq = fq.read().rstrip().decode('uft-16').split('\n')
                    fq = fq.read().rstrip().split('\n')
                    if fq != ['']:
                        for i in range(len(fq)):
                            if i%4 == 0:
                                newName = '>' + fq[i][1:]
                                rowSeq = fq[i+1]
                                fa.write(newName + '.2' + '\n' + rowSeq + '\n')
                            i += 1
            #merge unassembled reads
            os.system(f'cat {pearOutputPath}/{site}.1.fa {pearOutputPath}/{site}.2.fa > {pearOutputPath}/{site}.unassembled.fa')

            #change assembled reads to be fasta
            with open(f'{pearOutputPath}/{site}.assembled.fastq',encoding='utf8',errors='ignore') as fq:
                with open(f'{pearOutputPath}/{site}.assembled.fa','w') as fa:
                    #fqList = fq.read().rstrip().decode('uft-16').split('\n')
                    fq = fq.read().rstrip().split('\n')
                    if fq != ['']:
                        i = 0
                        for i in range(len(fq)):
                            if i%4 == 0:
                                newName = '>' + fq[i][1:]
                                rowSeq = fq[i+1]
                                fa.write(newName + '\n' + rowSeq + '\n')
                            i += 1
            #merge all reads
            os.system(f'cat {pearOutputPath}/{site}.assembled.fa {pearOutputPath}/{site}.unassembled.fa > {pearOutputPath}/{site}.all.fa')

    return None


##############
#Function: generate bash script for CAP3
#out_dir:output directory for assemble
#return: path of script for CAP3
def cap3(out_dir):
    listSites = os.listdir(out_dir)
    with open(f'{out_dir}/cap3.sh','w') as bashFile:
        bashFile.write('#!/bin/bash\n')
        for site in listSites:
            if ".sh" not in site:
                faPath = f'{out_dir}/{site}/pearOutput/'
                cap3OutputPath = f'{out_dir}/{site}/cap3Output/'
                mkdir(cap3OutputPath)
                bashFile.write(f'''
        cap3 {faPath}/{site}.all.fa > {cap3OutputPath}/{site}.CAP3
        ''')
        bashFile.write('echo \'CAP3 done\'')
    return f'{out_dir}/cap3.sh'